@@ -72,6 +72,7 @@ int getrandom(WORD08 *out,WORD32 len,VPN *anchor)
  *
  * input:  data   - the data the checksum is to be created for
  *	   length - the length of the data
+ *	   unused - unused, for hmac compatability
  *
  * output: out - the location the 2 byte checksum is to be stored at
  *
@@ -81,7 +82,7 @@ int getrandom(WORD08 *out,WORD32 len,VPN *anchor)
 
 #ifndef CRYPTOX86
 
-void checksum(register WORD08 *data,register WORD32 length,WORD08 *out)
+void checksum(register WORD08 *data,register WORD32 length,WORD08 *out,void *unused)
 {
 register WORD32 sum=0;	/* checksum creation buffer */
 
@@ -296,14 +297,13 @@ void decrypt_cfb_64_8(WORD08 *data,WORD32 total,VPN *anchor)
 int datasend(register WORD08 *data,register WORD32 length,VPN *anchor)
 {
 	register WORD32 i;			/* counter/index	*/
-	WORD08 hdr[4];				/* data header		*/
+	WORD08 bfr[22+MAXBLOCK+(MAXBLOCK/2)];	/* work buffer		*/
+	WORD08 *hdr=bfr+20-anchor->mac;		/* data header		*/
+	WORD08 *tmp=bfr+22;			/* compression/data bfr	*/
+						/* new key and iv memory*/
 #ifdef COMPRESS
 	WORD32 tmplen;				/* used for compression	*/
-	WORD08 tmp[MAXBLOCK+(MAXBLOCK/2)];	/* compression buffer	*/
 #endif
-						/* new key and iv memory*/
-	WORD08 key[BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY];
-	WORD08 keymem[BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY];
 
 
 	/* debug message */
@@ -324,7 +324,7 @@ int datasend(register WORD08 *data,register WORD32 length,VPN *anchor)
 
 		/* create length header */
 
-		hdr[2]=hdr[3]=0;
+		hdr[anchor->mac]=hdr[anchor->mac+1]=0;
 
 		/* server */
 
@@ -332,40 +332,39 @@ int datasend(register WORD08 *data,register WORD32 length,VPN *anchor)
 		{
 			/* get 80 random bytes */
 
-			if(!getrandom(key,BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY,
+			if(!getrandom(tmp,BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY,
 				anchor))GOTO("getrandom",keyerr);
 
 			/* encrypt 80 bytes in ECB mode (you never know) */
 
 			for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;
 				i+=BLOWFISH_BLOCKSIZE)
-				blowfish_encrypt(key+i,anchor->local);
+				blowfish_encrypt(tmp+i,anchor->local);
 
 			/* create checksum header */
 
-			checksum(key,BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY,hdr);
+			(anchor->macproc)(tmp,BLOWFISH_BLOCKSIZE
+					+BLOWFISH_MAXKEY,
+					hdr,(void *)(&(anchor->hmkey)));
 
 			/* memorize new key and iv */
 
 			for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
-				keymem[i]=key[i];
+			    tmp[i+BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY]=tmp[i];
 
 			/* encrypt header */
 
-			encrypt_cfb_64_8(hdr,4,anchor);
-
-			/* send header to peer, handle errors */
-
-			if(!dosend(hdr,4,anchor))GOTO("dosend hdr",keyerr);
+			encrypt_cfb_64_8(hdr,anchor->mac+2,anchor);
 
 			/* encrypt data */
 
-			encrypt_cfb_64_8(key,
+			encrypt_cfb_64_8(tmp,
 				BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY,anchor);
 
 			/* send data to peer, handle errors */
 
-			if(!dosend(key,BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY,
+			if(!dosend(hdr,BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY+
+				anchor->mac+2,
 				anchor))GOTO("dosend key",keyerr);
 
 			/* toggle local key schedule */
@@ -376,7 +375,8 @@ int datasend(register WORD08 *data,register WORD32 length,VPN *anchor)
 
 			/* generate schedule from new key */
 
-			blowfish_key_schedule(keymem+BLOWFISH_BLOCKSIZE,
+			blowfish_key_schedule(tmp+BLOWFISH_BLOCKSIZE
+				+BLOWFISH_MAXKEY+BLOWFISH_BLOCKSIZE,
 				anchor->keysize,anchor->local);
 		}
 
@@ -386,36 +386,34 @@ int datasend(register WORD08 *data,register WORD32 length,VPN *anchor)
 		{
 			/* get 8 random bytes */
 
-			if(!getrandom(key,BLOWFISH_BLOCKSIZE,anchor))
+			if(!getrandom(tmp,BLOWFISH_BLOCKSIZE,anchor))
 				GOTO("getrandom",keyerr);
 
 			/* encrypt 8 bytes in ECB mode (you never know) */
 
-			blowfish_encrypt(key,anchor->local);
+			blowfish_encrypt(tmp,anchor->local);
 
 			/* create checksum header */
 
-			checksum(key,BLOWFISH_BLOCKSIZE,hdr);
+			(anchor->macproc)(tmp,BLOWFISH_BLOCKSIZE,hdr,
+					(void *)(&(anchor->hmkey)));
 
 			/* memorize new iv */
 
-			for(i=0;i<BLOWFISH_BLOCKSIZE;i++)keymem[i]=key[i];
+			for(i=0;i<BLOWFISH_BLOCKSIZE;i++)
+			    tmp[i+BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY]=tmp[i];
 
 			/* encrypt header */
 
-			encrypt_cfb_64_8(hdr,4,anchor);
-
-			/* send header to peer, handle errors */
-
-			if(!dosend(hdr,4,anchor))GOTO("dosend hdr",keyerr);
+			encrypt_cfb_64_8(hdr,anchor->mac+2,anchor);
 
 			/* encrypt data */
 
-			encrypt_cfb_64_8(key,BLOWFISH_BLOCKSIZE,anchor);
+			encrypt_cfb_64_8(tmp,BLOWFISH_BLOCKSIZE,anchor);
 
 			/* send data to peer, handle errors */
 
-			if(!dosend(key,BLOWFISH_BLOCKSIZE,anchor))
+			if(!dosend(hdr,BLOWFISH_BLOCKSIZE+anchor->mac+2,anchor))
 				GOTO("dosend key",keyerr);
 
 			/* from now on use key schedule of key received
@@ -426,12 +424,13 @@ int datasend(register WORD08 *data,register WORD32 length,VPN *anchor)
 
 		/* setup new iv */
 
-		for(i=0;i<BLOWFISH_BLOCKSIZE;i++)anchor->local_iv[i]=keymem[i];
+		for(i=0;i<BLOWFISH_BLOCKSIZE;i++)
+		  anchor->local_iv[i]=tmp[i+BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY];
 
 		/* clean up */
 
 		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
-			key[i]=keymem[i]=0;
+			tmp[i]=tmp[i+BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY]=0;
 
 		/* debug message */
 
@@ -444,7 +443,7 @@ int datasend(register WORD08 *data,register WORD32 length,VPN *anchor)
 		/* error handling */
 
 keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
-			key[i]=keymem[i]=0;
+			tmp[i]=tmp[i+BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY]=0;
 		return 0;
 	}
 
@@ -499,61 +498,63 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 
 		DEBUG1(3,"compressed data\n");
 
-		/* create length header with MSB set */
-
-		hdr[2]=(WORD08)(tmplen);
-		hdr[3]=((WORD08)(tmplen>>8))|0x80;
-
-		/* set data pointer to compressed data buffer */
-
-		data=tmp;
-
 		/* change data length to compressed data length */
 
 		length=tmplen;
+
+		/* create length header with MSB set */
+
+		hdr[anchor->mac]=(WORD08)(length);
+		hdr[anchor->mac+1]=((WORD08)(length>>8))|0x80;
 	}
 
 	/* in case of no compression */
 
 	else
 	{
+		/* copy data to buffer */
+
+		memcpy(tmp,data,length);
+
 		/* create regular length header */
 
-		hdr[2]=(WORD08)(length);
-		hdr[3]=(WORD08)(length>>8);
+		hdr[anchor->mac]=(WORD08)(length);
+		hdr[anchor->mac+1]=(WORD08)(length>>8);
 	}
 #else
+	/* copy data to buffer */
+
+	memcpy(tmp,data,length);
+
 	/* create length header */
 
-	hdr[2]=(WORD08)(length);
-	hdr[3]=(WORD08)(length>>8);
+	hdr[anchor->mac]=(WORD08)(length);
+	hdr[anchor->mac+1]=(WORD08)(length>>8);
 #endif
 
 	/* create checksum header */
 
-	checksum(data,length,hdr);
+	(anchor->macproc)(tmp,length,hdr,(void *)(&(anchor->hmkey)));
 
-	/* debug message */
+	/* debug message (only the first four bytes of checksum for hmac) */
 
 	DEBUG3(4,"length=%04lx chksum=%04lx\n",
-		((WORD32)(hdr[3])<<8)|(WORD32)(hdr[2]),
+		((WORD32)(hdr[anchor->mac+1])<<8)|(WORD32)(hdr[anchor->mac]),
+		(anchor->hmacmode?((WORD32)(hdr[3])<<24)|
+		(WORD32)(hdr[2]<<16):0)|
 		((WORD32)(hdr[1])<<8)|(WORD32)(hdr[0]));
 
 	/* encrypt header if not in plaintext mode */
 
-	encrypt_cfb_64_8(hdr,4,anchor);
-
-	/* send header to peer, handle errors */
-
-	if(!dosend(hdr,4,anchor))GOTO("dosend hdr",keyerr);
+	encrypt_cfb_64_8(hdr,anchor->mac+2,anchor);
 
 	/* encrypt data if not in plaintext mode */
 
-	encrypt_cfb_64_8(data,length,anchor);
+	encrypt_cfb_64_8(tmp,length,anchor);
 
-	/* send data to peer, handle errors */
+	/* send header and data to peer, handle errors */
 
-	if(!dosend(data,length,anchor))GOTO("dosend data",keyerr);
+	if(!dosend(hdr,length+anchor->mac+2,anchor))GOTO("dosend data",keyerr);
 
 	/* debug message */
 
@@ -588,16 +589,19 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 
 int datarecv(register WORD08 *data,WORD32 max,VPN *anchor)
 {
-	WORD08 hdr[4];				/* data header		*/
-	WORD08 sum[2];				/* calculated checksum	*/
+	WORD08 hdr[22];				/* data header		*/
+	WORD08 sum[20];				/* calculated checksum	*/
 	register WORD32 i;			/* counter/index	*/
 	register WORD32 length;			/* data length		*/
 #ifdef COMPRESS
 	WORD32 tmplen=max;			/* used for compression	*/
 	WORD08 tmp[MAXBLOCK+(MAXBLOCK/2)];	/* compression buffer	*/
-#endif
+						/* new key and iv 	*/
+	WORD08 *newkey=tmp;
+#else
 						/* new key and iv 	*/
 	WORD08 newkey[BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY];
+#endif
 
 
 	/* debug message */
@@ -606,21 +610,22 @@ int datarecv(register WORD08 *data,WORD32 max,VPN *anchor)
 
 	/* read header from peer, signal errors */
 
-	if(!dorecv(hdr,4,anchor))RET("dorecv hdr",0);
+	if(!dorecv(hdr,anchor->mac+2,anchor))RET("dorecv hdr",0);
 
 	/* decrypt received data if not in plaintext mode */
 
-	decrypt_cfb_64_8(hdr,4,anchor);
+	decrypt_cfb_64_8(hdr,anchor->mac+2,anchor);
 
-	/* debug message */
+	/* debug message (only first four bytes for hmac) */
 
 	DEBUG3(4,"length=%04lx chksum=%04lx\n",
-		((WORD32)(hdr[3])<<8)|(WORD32)(hdr[2]),
+		((WORD32)(hdr[anchor->mac+1])<<8)|(WORD32)(hdr[anchor->mac]),
+		(anchor->hmacmode?((WORD32)(hdr[3])<<24)|(WORD32)(hdr[2]<<16):0)|
 		((WORD32)(hdr[1])<<8)|(WORD32)(hdr[0]));
 
 	/* new key from peer */
 
-	if(!hdr[2]&&!hdr[3])
+	if(!hdr[anchor->mac]&&!hdr[anchor->mac+1])
 	{
 		/* illegal, if in plaintext mode */
 
@@ -645,17 +650,21 @@ int datarecv(register WORD08 *data,WORD32 max,VPN *anchor)
 
 			/* calculate checksum of read data */
 
-			checksum(newkey,BLOWFISH_BLOCKSIZE,sum);
+			(anchor->macproc)(newkey,BLOWFISH_BLOCKSIZE,sum,
+				(void *)(&(anchor->hmkey)));
 
 			/* debug message */
 
 			DEBUG3(4,"length=%04lx chksum=%04lx\n",
-				((WORD32)(hdr[3])<<8)|(WORD32)(hdr[2]),
-				((WORD32)(hdr[1])<<8)|(WORD32)(hdr[0]));
+				((WORD32)(hdr[anchor->mac+1])<<8)|
+					(WORD32)(hdr[anchor->mac]),
+				(anchor->hmacmode?((WORD32)(hdr[3])<<24)|
+					(WORD32)(hdr[2]<<16):0)|
+					((WORD32)(hdr[1])<<8)|(WORD32)(hdr[0]));
 
 			/* in case of checksum mismatch signal error */
 
-			if(hdr[0]!=sum[0]||hdr[1]!=sum[1])
+			for(i=0;i<anchor->mac;i++)if(hdr[i]!=sum[i])
 				GOTO("chksum error",keyerr);
 
 			/* from now on use key schedule 
@@ -678,18 +687,21 @@ int datarecv(register WORD08 *data,WORD32 max,VPN *anchor)
 
 			/* calculate checksum of read data */
 
-			checksum(newkey,BLOWFISH_BLOCKSIZE+
-				BLOWFISH_MAXKEY,sum);
+			(anchor->macproc)(newkey,BLOWFISH_BLOCKSIZE+
+				BLOWFISH_MAXKEY,sum,(void *)(&(anchor->hmkey)));
 
 			/* debug message */
 
 			DEBUG3(4,"length=%04lx chksum=%04lx\n",
-				((WORD32)(hdr[3])<<8)|(WORD32)(hdr[2]),
-				((WORD32)(hdr[1])<<8)|(WORD32)(hdr[0]));
+				((WORD32)(hdr[anchor->mac+1])<<8)|
+					(WORD32)(hdr[anchor->mac]),
+				(anchor->hmacmode?((WORD32)(hdr[3])<<24)|
+					(WORD32)(hdr[2])<<16:0)|
+					((WORD32)(hdr[1])<<8)|(WORD32)(hdr[0]));
 
 			/* in case of checksum mismatch signal error */
 
-			if(hdr[0]!=sum[0]||hdr[1]!=sum[1])
+			for(i=0;i<anchor->mac;i++)if(hdr[i]!=sum[i])
 				GOTO("chksum error",keyerr);
 
 			/* toggle remote key schedule */
@@ -731,9 +743,9 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 	/* restore data length */
 
 #ifdef COMPRESS
-	length=(((WORD32)((hdr[3]&0x7f)))<<8)|((WORD32)(hdr[2]));
+	length=(((WORD32)((hdr[anchor->mac+1]&0x7f)))<<8)|((WORD32)(hdr[anchor->mac]));
 #else
-	length=(((WORD32)(hdr[3]))<<8)|((WORD32)(hdr[2]));
+	length=(((WORD32)(hdr[anchor->mac+1]))<<8)|((WORD32)(hdr[anchor->mac]));
 #endif
 	/* signal error if length exceeds buffer size */
 
@@ -743,7 +755,7 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 
 	/* in case of compressed data */
 
-	if(hdr[3]&0x80)
+	if(hdr[anchor->mac+1]&0x80)
 	{
 		/* debug message */
 
@@ -759,7 +771,8 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 
 		/* calculate checksum of read data */
 
-		checksum(tmp,length,sum);
+		(anchor->macproc)(tmp,length,sum,
+			(void *)(&(anchor->hmkey)));
 
 		/* decompress data, signal errors */
 
@@ -785,12 +798,13 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 
 		/* calculate checksum of read data */
 
-		checksum(data,length,sum);
+		(anchor->macproc)(data,length,sum,
+			(void *)(&(anchor->hmkey)));
 	}
 #else
 	/* read expected amount of data, signal errors */
 
-	if(!dorecv(data,length))RET("dorecv data",0);
+	if(!dorecv(data,length,anchor))RET("dorecv data",0);
 
 	/* decrypt received data if not in plaintext mode */
 
@@ -798,7 +812,7 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 
 	/* calculate checksum of read data */
 
-	checksum(data,length,sum);
+	(anchor->macproc)(data,length,sum,(void *)(&(anchor->hmkey)));
 #endif
 	/* debug message */
 
@@ -810,14 +824,20 @@ keyerr:		for(i=0;i<BLOWFISH_BLOCKSIZE+BLOWFISH_MAXKEY;i++)
 		for(l=0;l<length;l++)printf("%02x ",data[l]);
 		printf("\n");
 	}
+	if(debug>3)
+	{
+		int l;
+		printf("chksum=");
+		for(l=0;l<anchor->mac;l++)printf("%02x",hdr[l]);
+		printf(" expected=");
+		for(l=0;l<anchor->mac;l++)printf("%02x",sum[l]);
+		printf("\n");
+	}
 #endif
-	DEBUG3(4,"chksum=%04lx expected=%04lx\n",
-		((WORD32)(hdr[1])<<8)|(WORD32)(hdr[0]),
-		((WORD32)(sum[1])<<8)|(WORD32)(sum[0]));
 
 	/* in case of checksum mismatch signal error */
 
-	if(hdr[0]!=sum[0]||hdr[1]!=sum[1])RET("chksum error",0);
+	for(i=0;i<anchor->mac;i++)if(hdr[i]!=sum[i])RET("chksum error",0);
 
 	/* debug message */
 
@@ -944,13 +964,12 @@ void genwhite(WORD08 *src,VPN *anchor)
 	/* encrypt the buffer four times */
 
 	for(i=0;i<256;i+=BLOWFISH_BLOCKSIZE)
+	{
 		blowfish_encrypt(anchor->white+i,sched);
-	for(i=0;i<256;i+=BLOWFISH_BLOCKSIZE)
 		blowfish_encrypt(anchor->white+i,sched);
-	for(i=0;i<256;i+=BLOWFISH_BLOCKSIZE)
 		blowfish_encrypt(anchor->white+i,sched);
-	for(i=0;i<256;i+=BLOWFISH_BLOCKSIZE)
 		blowfish_encrypt(anchor->white+i,sched);
+	}
 
 	/* clean up */
 
@@ -1302,6 +1321,7 @@ crterr:		printf("Can't create %s\n",anchor->keyfile);
  *	  15 - challenge receive error
  *	  16 - challenge comparison error
  *	  17 - master key file save error
+ *	  18 - peer doesn't support hmac
  *
  * This procedure establishes encrypted communication with
  * the remote system either in server or in client mode.
@@ -1313,7 +1333,7 @@ int cryptosetup(VPN *anchor)
 	WORD32 src;			/* numeric of local ip		*/
 	WORD32 dst;			/* numeric of remote ip		*/
 	WORD08 xbfr[32];		/* extended mode data buffer	*/
-	WORD08 id=0;			/* selected transaction id	*/
+	WORD32 id=0;			/* selected transaction id	*/
 	WORD08 *secret;			/* shared secret memory area	*/
 	WORD08 *master;			/* current master secret	*/
 	WORD08 *iv;			/* current master secret iv	*/
@@ -1356,6 +1376,10 @@ int cryptosetup(VPN *anchor)
 #ifdef COMPRESS
 			if(!anchor->nocompress)xbfr[7]|=1;
 #endif
+			/* set up hmac capabilities */
+
+			xbfr[7]|=2|((WORD08)((anchor->hmacmode)<<2));
+
 			/* copy key id to transfer buffer */
 
 			memcpy(xbfr+8,secret+KEYID,24);
@@ -1395,6 +1419,10 @@ int cryptosetup(VPN *anchor)
 #ifdef COMPRESS
 			if(!anchor->nocompress)xbfr[15]|=1;
 #endif
+			/* set up hmac capabilities */
+
+			xbfr[15]|=2|((WORD08)((anchor->hmacmode)<<2));
+
 			/* send buffer to peer, handle errors */
 
 			if(!dosend(xbfr+8,8,anchor))RET("dosend",6);
@@ -1425,6 +1453,27 @@ int cryptosetup(VPN *anchor)
 		if(xbfr[7]&1)anchor->peercomp=0;
 		else anchor->peercomp=1;
 #endif
+
+		/* handle hmac options */
+
+		if(!(xbfr[7]&2))
+		{
+			if(anchor->hmacopt==2)
+			{
+				RET("peer doesn't support hmac",18);
+			}
+			else i=0;
+		}
+		else
+		{
+			i=(xbfr[7]&0xc)>>2;
+			i=(i>=anchor->hmacmode?i:anchor->hmacmode);
+		}
+
+		anchor->mackey=macinit[i];
+		anchor->macproc=mactab[i];
+		anchor->mac=maclen[i];
+
 		/* send remote iv to peer, handle errors */
 
 		if(!dosend(iv,BLOWFISH_BLOCKSIZE,anchor))RET("dosend",9);
@@ -1493,6 +1542,11 @@ int cryptosetup(VPN *anchor)
 
 	blowfish_key_schedule(master,anchor->keysize,anchor->u1.sched);
 
+	/* create preprocessed hmac key in hmac mode */
+
+	if(anchor->mackey)(anchor->mackey)(master,BLOWFISH_MAXKEY,
+		(void *)(&(anchor->hmkey)));
+
 	/* in client mode */
 
 	if(anchor->mode)
@@ -1594,7 +1648,7 @@ int cryptosetup(VPN *anchor)
 
 		/* write new transaction id to master key buffer */
 
-		secret[TRANSACTID]=id;
+		secret[TRANSACTID]=(WORD08)(id);
 
 		/* save master key file, handle errors */
 
