@@ -79,6 +79,9 @@ static CONFIG config[]=
 { "ipopts",    1,1,1 },
 { "sendbuf",   1,1,1 },
 { "connwait",  1,1,1 },
+{ "hmac",      1,1,2 },
+{ "slipup",    1,1,1 },
+{ "slipdown",  1,1,1 },
 { NULL,        0,0,0 }
 };
 
@@ -118,6 +121,9 @@ static CONFIG config[]=
 #define IDX_IPOPTS		40
 #define IDX_SENDBUF		41
 #define IDX_CONNWAIT		42
+#define IDX_HMAC		43
+#define IDX_SLIPUP		44
+#define IDX_SLIPDOWN		45
 
 #define CFG_MODE(a)		anchor->config[IDX_MODE].a
 #define CFG_LOCAL(a)		anchor->config[IDX_LOCAL].a
@@ -156,6 +162,9 @@ static CONFIG config[]=
 #define CFG_IPOPTS(a)		anchor->config[IDX_IPOPTS].a
 #define CFG_SENDBUF(a)		anchor->config[IDX_SENDBUF].a
 #define CFG_CONNWAIT(a)		anchor->config[IDX_CONNWAIT].a
+#define CFG_HMAC(a)		anchor->config[IDX_HMAC].a
+#define CFG_SLIPUP(a)		anchor->config[IDX_SLIPUP].a
+#define CFG_SLIPDOWN(a)		anchor->config[IDX_SLIPDOWN].a
 
 /* numeric configuration data keep sequence! */
 
@@ -192,6 +201,7 @@ static NUMDATA numdata[]=
 	{ IDX_NOCOMPRESS, -1, 1,               0               },
 	{ IDX_XFILTER,    -1, 1,               0               },
 	{ IDX_OOBFIX,     -1, 1,               0               },
+	{ IDX_HMAC,        1, 2,               0               },
 	{ 0,               0, 0,               0               }
 
 };
@@ -244,7 +254,7 @@ int checkfile(char *pathname,int perms)
 
 	if(lstat(pathname,&stb))
 	{
-		printf("can't stat '%s'\n",pathname);
+		printf("Can't stat '%s'.\n",pathname);
 		return 0;
 	}
 
@@ -256,12 +266,12 @@ int checkfile(char *pathname,int perms)
 
 		if(!S_ISCHR(stb.st_mode))
 		{
-			printf("'%s' is not a character special device\n",
+			printf("'%s' is not a character special device.\n",
 				pathname);
 			return 0;
 		}
 
-		if(!stb.st_mode)printf("'%s' has permission 0\n",pathname);
+		if(!stb.st_mode)printf("'%s' has permission 0.\n",pathname);
 
 		/* signal success or error */
 
@@ -272,7 +282,7 @@ int checkfile(char *pathname,int perms)
 
 	if(!S_ISREG(stb.st_mode))
 	{
-		printf("'%s': not a regular file\n",pathname);
+		printf("'%s': not a regular file.\n",pathname);
 		return 0;
 	}
 
@@ -280,7 +290,7 @@ int checkfile(char *pathname,int perms)
 
 	if(stb.st_uid||stb.st_gid)
 	{
-		printf("'%s' must be owned by root:root\n",pathname);
+		printf("'%s' must be owned by root:root.\n",pathname);
 		return 0;
 	}
 
@@ -288,12 +298,12 @@ int checkfile(char *pathname,int perms)
 
 	if(stb.st_mode&perms)
 	{
-		printf("'%s' permissions must be %03o or better",
+		printf("'%s' permissions must be %03o or better.\n",
 			pathname,0777&perms);
 		return 0;
 	}
 
-	if(!stb.st_size)printf("'%s' has size 0\n",pathname);
+	if(!stb.st_size)printf("'%s' has size 0.\n",pathname);
 
 	/* return file size (0=error) */
 
@@ -307,7 +317,7 @@ int checkfile(char *pathname,int perms)
  *
  * return: 1 in case of valid ip, else 0
  *
- * This procedure chechs the given ip for validity
+ * This procedure checks the given ip for validity
  */
 
 int chkip(char *ip)
@@ -356,7 +366,7 @@ int chksvrcln(char mode,char *p1,char *p2,char *txt,VPN *anchor)
 			strcpy(mode?anchor->client_ip:anchor->server_ip,p1);
 		else
 		{
-			printf("illegal %s ip\n",txt);
+			printf("Illegal %s ip.\n",txt);
 			return 0;
 		}
 	}
@@ -367,12 +377,12 @@ int chksvrcln(char mode,char *p1,char *p2,char *txt,VPN *anchor)
 	{
 		if(strncmp(p1,"/dev/",5))
 		{
-			printf("illegal %s device (not in /dev)\n",txt);
+			printf("Illegal %s device (not in /dev).\n",txt);
 			return 0;
 		}
 		if(p2)
 		{
-			printf("do not use port for %s device\n",txt);
+			printf("Do not use port for %s device.\n",txt);
 			return 0;
 		}
 		if(anchor->mode==mode)
@@ -461,7 +471,7 @@ int parse(VPN *anchor)
 
 			if(v.c->found)
 			{
-				printf("command '%s' repeated.\n",item);
+				printf("Command '%s' repeated.\n",item);
 				fclose(fp);
 				return 0;
 			}
@@ -487,7 +497,7 @@ int parse(VPN *anchor)
 
 				if(j==u.c->max)
 				{
-					printf("command '%s' takes only %d " \
+					printf("Command '%s' takes only %d " \
 						"parameters.\n",u.c->item,
 						u.c->max);
 					fclose(fp);
@@ -512,7 +522,7 @@ int parse(VPN *anchor)
 
 			if(j<u.c->min)
 			{
-				printf("command '%s' takes at least %d " \
+				printf("Command '%s' takes at least %d " \
 					"parameters.\n",u.c->item,u.c->min);
 				fclose(fp);
 				return 0;
@@ -527,7 +537,7 @@ int parse(VPN *anchor)
 
 		if(!u.c->item)
 		{
-			printf("illegal command '%s' encountered.\n",item);
+			printf("Illegal command '%s' encountered.\n",item);
 			fclose(fp);
 			return 0;
 		}
@@ -542,7 +552,7 @@ int parse(VPN *anchor)
 	for(u.c=config,v.c=anchor->config;u.c->item;u.c++,v.c++)
 		if(!u.c->option)if(!v.c->found)
 	{
-		printf("missing command '%s'\n",u.c->item);
+		printf("Missing command '%s'.\n",u.c->item);
 		return 0;
 	}
 
@@ -552,7 +562,7 @@ int parse(VPN *anchor)
 	else if(!strcmp(CFG_MODE(p1),"server"))anchor->mode=0;
 	else
 	{
-		printf("mode must be 'client' or 'server'\n");
+		printf("Mode must be 'client' or 'server'.\n");
 		return 0;
 	}
 
@@ -561,7 +571,7 @@ int parse(VPN *anchor)
 	if(chkip(CFG_LOCAL(p1)))strcpy(anchor->local_ip,CFG_LOCAL(p1));
 	else
 	{
-		printf("illegal local ip\n");
+		printf("Illegal local ip.\n");
 		return 0;
 	}
 
@@ -570,7 +580,7 @@ int parse(VPN *anchor)
 	if(chkip(CFG_REMOTE(p1)))strcpy(anchor->remote_ip,CFG_REMOTE(p1));
 	else
 	{
-		printf("illegal remote ip\n");
+		printf("Illegal remote ip.\n");
 		return 0;
 	}
 
@@ -584,7 +594,7 @@ int parse(VPN *anchor)
 	if((CFG_SERVER(p1[0])=='/'&&CFG_CLIENT(p1[0])!='/')||
 		(CFG_SERVER(p1[0])!='/'&&CFG_CLIENT(p1[0])=='/'))
 	{
-		printf("use either ip or device for both client and server\n");
+		printf("Use either ip or device for both client and server.\n");
 		return 0;
 	}
 
@@ -654,7 +664,7 @@ int parse(VPN *anchor)
 
 			/* signal error if value is out of range */
 
-			printf("illegal %s value, valid range is %d to %d\n",
+			printf("Illegal %s value, valid range is %d to %d.\n",
 				config[j].item,u.n->lower,u.n->upper);
 			return 0;
 		}
@@ -670,7 +680,7 @@ int parse(VPN *anchor)
 	anchor->mtu=*u.w16++;
 	anchor->retry=*u.w16++;
 	anchor->noresponse=*u.w16++;
-	anchor->keysize=*u.w16++;
+	anchor->keysize=(WORD08)(*u.w16++);
 	anchor->threshold=*u.w16++;
 	anchor->linktest=*u.w16++;
 	anchor->suspend=*u.w16++;
@@ -686,6 +696,7 @@ int parse(VPN *anchor)
 	anchor->nocompress=(WORD08)(*u.w16++);
 	anchor->xfilter=(WORD08)(*u.w16++);
 	anchor->oobfix=(WORD08)(*u.w16++);
+	anchor->hmacopt=(int)(*u.w16++);
 
 	/* check all routes, if given */
 
@@ -694,8 +705,8 @@ int parse(VPN *anchor)
 		if(!chkip(CFG_ROUTE(i,p1))||!chkip(CFG_ROUTE(i,p2))||
 			!chkip(CFG_ROUTE(i,p3)))
 		{
-			printf("illegal destination, netmask or gateway " \
-				"for '%s' command\n",config[i].item);
+			printf("Illegal destination, netmask or gateway " \
+				"for '%s' command.\n",config[i].item);
 			return 0;
 		}
 		anchor->rtab[anchor->rmax][0]=inet_addr(CFG_ROUTE(i,p1));
@@ -708,6 +719,25 @@ int parse(VPN *anchor)
 
 	/* get keyfile if given */
 
+	if(CFG_HMAC(found))
+	{
+		if(CFG_HMAC(p2))
+		{
+			if(!strcmp(CFG_HMAC(p2),"md5"))anchor->hmacmode=1;
+			else if(!strcmp(CFG_HMAC(p2),"sha1"))anchor->hmacmode=2;
+			else if(!strcmp(CFG_HMAC(p2),"rmd160"))
+				anchor->hmacmode=3;
+			else
+			{
+				printf("unknown mode for hmac option.\n");
+				return 0;
+			}
+		}
+		else anchor->hmacmode=1;
+	}
+
+	/* get keyfile if given */
+
 	if(CFG_KEYFILE(found))anchor->keyfile=CFG_KEYFILE(p1);
 
 	/* use optional pid file */
@@ -726,6 +756,14 @@ int parse(VPN *anchor)
 
 	if(CFG_LINKDOWN(found))anchor->linkdown=CFG_LINKDOWN(p1);
 
+	/* set pathname of linkup process if requested */
+
+	if(CFG_SLIPUP(found))anchor->slipup=CFG_SLIPUP(p1);
+
+	/* set pathname of linkdown process if requested */
+
+	if(CFG_SLIPDOWN(found))anchor->slipdown=CFG_SLIPDOWN(p1);
+
 	/* use optional modem chat file, check file permissions, too */
 
 	if(CFG_MODEMCHAT(found))
@@ -744,7 +782,7 @@ int parse(VPN *anchor)
 		if(speeddata[i].speednum)anchor->speed=speeddata[i].speedval;
 		else
 		{
-			printf("illegal speed value\n");
+			printf("Illegal speed value.\n");
 			return 0;
 		}
 	}
@@ -778,7 +816,7 @@ int parse(VPN *anchor)
 
 		else if(!peerroute(anchor))
 		{
-			printf("Can't find route to peer\n");
+			printf("Can't find route to peer.\n");
 			return 0;
 		}
 	}
@@ -791,22 +829,36 @@ int parse(VPN *anchor)
 				break;
 	case XKEYSIZE:		anchor->extended=1;
 				break;
-	default:		printf("'%s' has wrong size\n",anchor->keyfile);
+	default:		printf("'%s' has wrong size.\n",
+					anchor->keyfile);
 	case 0:			return 0;
 	}
 
 	/* check that passphrase is given only for extended mode
-	   master key file */
+	   master key file and that optional hmac is used only in
+	   extended mode */
 
-	if(!anchor->extended&&anchor->passwd)
+	if(!anchor->extended)
 	{
-		printf("-p option is only valid for extended master" \
-			" key files\n");
-		return 0;
+		if(anchor->passwd)
+		{
+			printf("-p option is only valid for extended master" \
+				" key files.\n");
+			return 0;
+		}
+		if(anchor->hmacopt==1)
+		{
+			printf("Optional hmac is only valid for extended " \
+				"master key files.\n");
+			return 0;
+		}
 	}
 
 	/* sanity settings */
 
+	anchor->mackey=macinit[anchor->hmacmode];
+	anchor->macproc=mactab[anchor->hmacmode];
+	anchor->mac=maclen[anchor->hmacmode];
 	if(!anchor->serialmode)anchor->xfilter=0;
 	else anchor->linktest=0;
 	if(!anchor->keysize||anchor->mode)anchor->keyttl=0;
